#define _CRT_NON_CONFORMING_SWPRINTFS

#include <stdarg.h>
#include <stdio.h>

#include <windows.h>
#include <gdiplus.h>

#define trace printf

#define ok(a, ...) ok_(__LINE__, (a), __VA_ARGS__)
static void ok_(int line, int condition, const char *fmt, ...)
{
    if (!condition)
    {
        va_list args;

        printf("line %d: ", line);
        va_start(args, fmt);
        vprintf(fmt, args);
        va_end(args);

        exit(-1);
    }
}

static UINT width, height;

static const REAL mm_per_inch = 25.4f;
static const REAL point_per_inch = 72.0f;

static void set_rect_empty(RectF *rc)
{
    rc->X = 0.0;
    rc->Y = 0.0;
    rc->Width = 0.0;
    rc->Height = 0.0;
}

/* converts a given unit to its value in pixels */
static REAL units_to_pixels(REAL units, GpUnit unit, REAL dpi)
{
    switch (unit)
    {
    case UnitPixel:
    case UnitDisplay:
        return units;
    case UnitPoint:
        return units * dpi / point_per_inch;
    case UnitInch:
        return units * dpi;
    case UnitDocument:
        return units * dpi / 300.0f; /* Per MSDN */
    case UnitMillimeter:
        return units * dpi / mm_per_inch;
    default:
        ok(0, "Unsupported unit: %d\n", unit);
        return 0;
    }
}

/* converts value in pixels to a given unit */
static REAL pixels_to_units(REAL pixels, GpUnit unit, REAL dpi)
{
    switch (unit)
    {
    case UnitPixel:
    case UnitDisplay:
        return pixels;
    case UnitPoint:
        return pixels * point_per_inch / dpi;
    case UnitInch:
        return pixels / dpi;
    case UnitDocument:
        return pixels * 300.0f / dpi;
    case UnitMillimeter:
        return pixels * mm_per_inch / dpi;
    default:
        ok(0, "Unsupported unit: %d\n", unit);
        return 0;
    }
}

static void draw_text(HDC hdc, int x, int y, float scale)
{
    WCHAR buf[256];
    GpStatus status;
    GpGraphics *graphics;
    GpMatrix *matrix;
    GpPointF pt[2];
    REAL elems[6], resx, resy;
    BOOL result;
    GpFontFamily *family;
    GpFont *font;
    GpStringFormat *format;
    GpBrush *brush;
    RectF rect, bounds;
    LOGFONTW lf;
    int lines;

    trace("hdc %p, x %d, y %d, scale %f\n", hdc, x, y, scale);

    status = GdipCreateFromHDC(hdc, &graphics);
    ok(status == Ok, "got %d\n", status);

    status = GdipGetDpiX(graphics, &resx);
    ok(status == Ok, "got %d\n", status);
    status = GdipGetDpiY(graphics, &resy);
    ok(status == Ok, "got %d\n", status);
    trace("resx %f, resy %f\n", resx, resy);

    status = GdipSetPageUnit(graphics, UnitInch);
    ok(status == Ok, "got %d\n", status);

    status = GdipSetPageScale(graphics, scale);
    ok(status == Ok, "got %d\n", status);

    status = GdipCreateMatrix(&matrix);
    ok(status == Ok, "got %d\n", status);
    status = GdipGetWorldTransform(graphics, matrix);
    ok(status == Ok, "got %d\n", status);
    status = GdipGetMatrixElements(matrix, elems);
    ok(status == Ok, "got %d\n", status);
    trace("world transform: %f, %f, %f, %f, %f, %f\n",
          elems[0], elems[1], elems[2], elems[3], elems[4], elems[5]);
    status = GdipIsMatrixIdentity(matrix, &result);
    ok(status == Ok, "got %d\n", status);
    ok(result, "got %d\n", result);
    GdipDeleteMatrix(matrix);

    pt[0].X = 0.0;
    pt[0].Y = 0.0;
    pt[1].X = 1.0;
    pt[1].Y = 1.0;
    status = GdipTransformPoints(graphics, CoordinateSpaceDevice, CoordinateSpaceWorld, pt, 2);
//    status = GdipTransformPoints(graphics, CoordinateSpaceDevice, CoordinateSpacePage, pt, 2);
    ok(status == Ok, "got %d\n", status);
    trace("transformed points: %f,%f, %f,%f\n", pt[0].X, pt[0].Y, pt[1].X, pt[1].Y);

    status = GdipCreateFontFamilyFromName(L"Times New Roman", NULL, &family);
    ok(status == Ok, "got %d\n", status);

    status = GdipCreateFont(family, pixels_to_units(100, UnitPoint, resy), FontStyleRegular, UnitPoint, &font);
//    status = GdipCreateFont(family, pixels_to_units(100, UnitInch, resy), FontStyleRegular, UnitInch, &font);
//    status = GdipCreateFont(family, pixels_to_units(100, UnitPixel, resy), FontStyleRegular, UnitPixel, &font);
    ok(status == Ok, "got %d\n", status);

    status = GdipCreateStringFormat(0, 0, &format);
    ok(status == Ok, "got %d\n", status);
    status = GdipSetStringFormatFlags(format, 0);
    ok(status == Ok, "got %d\n", status);
//    status = GdipSetStringFormatFlags(format, StringFormatFlagsNoWrap);
    status = GdipSetStringFormatFlags(format, StringFormatFlagsNoWrap | StringFormatFlagsNoClip);
    ok(status == Ok, "got %d\n", status);
    status = GdipSetStringFormatTrimming(format, StringTrimmingNone);
    ok(status == Ok, "got %d\n", status);
    status = GdipSetStringFormatHotkeyPrefix(format, HotkeyPrefixNone);
    ok(status == Ok, "got %d\n", status);
    status = GdipSetStringFormatLineAlign(format, StringAlignmentNear);
    ok(status == Ok, "got %d\n", status);

    status = GdipStringFormatGetGenericTypographic(&format);
    ok(status == Ok, "got %d\n", status);

    status = GdipCreateSolidFill((ARGB)0xff000000, (GpSolidFill **)&brush);
    ok(status == Ok, "got %d\n", status);

//    status = GdipSetTextRenderingHint(graphics, TextRenderingHintSingleBitPerPixel);
//    ok(status == Ok, "got %d\n", status);

    rect.X = pixels_to_units((float)x, UnitInch, resx) / scale;
    rect.Y = pixels_to_units((float)y, UnitInch, resy) / scale;
    rect.Width = pixels_to_units((float)width, UnitInch, resx) / scale;
    rect.Height = pixels_to_units((float)height, UnitInch, resy) / scale;
    set_rect_empty(&bounds);
    status = GdipMeasureString(graphics, L"Hello", 5, font, &rect, format, &bounds, NULL, &lines);
    ok(status == Ok, "got %d\n", status);
    trace("bounds (units): %f,%f-%f,%f, (lines %d)\n", bounds.X, bounds.Y, bounds.Width, bounds.Height, lines);
    trace("bounds (pixels): %f,%f-%f,%f, (lines %d)\n",
        units_to_pixels(bounds.X, UnitInch, resx), units_to_pixels(bounds.Y, UnitInch, resy),
        units_to_pixels(bounds.Width, UnitInch, resx), units_to_pixels(bounds.Height, UnitInch, resy), lines);

    swprintf(buf, L"x %d, y %d, scale %f\n", x, y, scale);
    status = GdipDrawString(graphics, buf, lstrlenW(buf), font, &rect, format, brush);
    ok(status == Ok, "got %d\n", status);

    status = GdipSetPageUnit(graphics, UnitPoint);
    ok(status == Ok, "got %d\n", status);

    status = GdipGetLogFontW(font, graphics, &lf);
    ok(status == Ok, "got %d\n", status);
    trace("GdipGetLogFontW: %ws, %d\n", lf.lfFaceName, lf.lfHeight);

    status = GdipDeleteFont(font);
    ok(status == Ok, "got %d\n", status);
    status = GdipDeleteFontFamily(family);
    ok(status == Ok, "got %d\n", status);
    status = GdipDeleteGraphics(graphics);
    ok(status == Ok, "got %d\n", status);
}

static LRESULT CALLBACK main_wnd_proc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    switch(msg)
    {
        case WM_PAINT:
        {
            PAINTSTRUCT ps;

            BeginPaint(hwnd, &ps);

            draw_text(ps.hdc, 0, 0, 1.0f);
            draw_text(ps.hdc, 0, 100, 0.1f);
            draw_text(ps.hdc, 0, 200, 0.01f);
            draw_text(ps.hdc, 0, 300, 2.0f);
            draw_text(ps.hdc, 0, 400, 10.0f);
            draw_text(ps.hdc, 0, 500, 100.0f);

            EndPaint(hwnd, &ps);
            return 0;
        }

        case WM_SIZE:
            width = LOWORD(lp);
            height = HIWORD(lp);
            break;

        case WM_DESTROY:
            PostQuitMessage(0);
            break;
    }

    return DefWindowProc(hwnd, msg, wp, lp);
}

static void create_window(void)
{
    WNDCLASSEX cls;

    memset(&cls, 0, sizeof(cls));
    cls.cbSize = sizeof(cls);
    cls.style = 0;
    cls.lpfnWndProc = main_wnd_proc;
    cls.hInstance = NULL;
    cls.hCursor = LoadCursor(0, IDC_ARROW);
    cls.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    cls.lpszClassName = "main";

    RegisterClassEx(&cls);
   
    CreateWindowEx(0, "main", NULL, WS_OVERLAPPEDWINDOW | WS_VISIBLE,
		   CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		   NULL, NULL, NULL, NULL);
}
 
int main(void)
{
    struct GdiplusStartupInput input;
    ULONG_PTR token;
    MSG msg;

    input.GdiplusVersion = 1;
    input.DebugEventCallback = NULL;
    input.SuppressBackgroundThread = FALSE;
    input.SuppressExternalCodecs = TRUE;
    GdiplusStartup(&token, &input, NULL);

    create_window();
    
    while (GetMessage(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    GdiplusShutdown(token);

    return 0;
}
